@@ -14,7 +14,7 @@
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
 
-/* Steven Schubiger <schubiger@gmail.com> */
+/* Steven Schubiger <stsc@members.fsf.org> */
 
 #include <config.h>
 
@@ -33,6 +33,7 @@
 #include "binary-io.h"
 #include "closeout.h"
 #include "error.h"
+#include "full-write.h"
 #include "quote.h"
 #include "safe-read.h"
 #include "xalloc.h"
@@ -59,13 +60,13 @@
 #define PROGRAM_NAME "linecut"
 
 /* The version of this program (floating-point value). */
-#define PROGRAM_VERSION 0.3
+#define PROGRAM_VERSION 0.4
 
 /* The authors of this program. */
 #define AUTHORS "Steven Schubiger"
 
 /* The address to send bug-reports to. */
-#define BUGS_MAIL_ADDRESS "schubiger@gmail.com"
+#define BUGS_MAIL_ADDRESS "stsc@members.fsf.org"
 
 /* Maximum of allowed range sets. */
 #define MAX_RANGE_SETS (16 * 2)
@@ -206,9 +207,9 @@ Report bugs to <%s>.\n\
   exit (status);
 }
 
-static void 
+static void
 version (void)
-{      
+{
   printf (_("\
 %s %.1f\n\
 "), PROGRAM_NAME, PROGRAM_VERSION);
@@ -259,7 +260,7 @@ elide_lines_seekable (const char *filename, int fd,
                       off_t offset_from, off_t offset_to)
 {
   char buffer[BUFSIZ];
-  char *buf, *next, *curr;
+  char *buf, *eol;
   bool seen_complete_line;
   uintmax_t line_current;
 
@@ -274,7 +275,7 @@ elide_lines_seekable (const char *filename, int fd,
   while (1)
     {
       size_t bytes_read = safe_read (fd, buffer, BUFSIZ);
-      size_t bytes_to_write = 0;
+      size_t bytes_to_write;
 
       if (bytes_read == SAFE_READ_ERROR)
         {
@@ -284,21 +285,18 @@ elide_lines_seekable (const char *filename, int fd,
       if (bytes_read == 0)
         break;
       buf = buffer;
+      bytes_to_write = bytes_read;
       for (; line_current <= line_to; line_current++)
         {
-          if (bytes_to_write < bytes_read)
+          if (bytes_to_write)
             {
               /* full line with trailing newline detected. */
-              if (strstr (buf, "\n") != NULL)
+              if (eol = memchr (buf, '\n', bytes_to_write))
                 {
-                  curr = (char *) rawmemchr (buf, '\n');
-                  next = curr;
-                  next++;
-                  *curr = '\0';
-
                   if (number_mode && seen_complete_line)
                     {
-                      if (fwrite (line_num_print, 1, strlen (line_num_print), stdout) < strlen (line_num_print))
+                      if (full_write (STDOUT_FILENO, line_num_print, strlen (line_num_print)) 
+                          != strlen (line_num_print))
                         {
                           error (0, errno, _("write error"));
                           return false;
@@ -307,38 +305,38 @@ elide_lines_seekable (const char *filename, int fd,
                     }
                   seen_complete_line = true;
 
-                  if (fwrite (buf, 1, curr - buf, stdout) < curr - buf)
+                  if (full_write (STDOUT_FILENO, buf, eol - buf) != eol - buf)
                     {
                       error (0, errno, _("write error"));
                       return false;
                     }
-                  if (fwrite ("\n", 1, 1, stdout) < 1)
+                  if (full_write (STDOUT_FILENO, "\n", 1) != 1)
                     {
                       error (0, errno, _("write error"));
                       return false;
                     }
 
-                  bytes_to_write += curr - buf;
-                  buf = next;
+                  bytes_to_write -= eol - buf + 1; /* + 1 for subtracting the newline too. */
+                  buf = eol + 1; /* + 1 because we advance beyond the end-of-line. */
                 }
               /* part of line with no trailing newline. */
               else
                 {
                   if (number_mode)
                     {
-                      if (fwrite (line_num_print, 1, strlen (line_num_print), stdout) < strlen (line_num_print))
+                      if (full_write (STDOUT_FILENO, line_num_print, strlen (line_num_print))
+                          != strlen (line_num_print))
                         {
                           error (0, errno, _("write error"));
                           return false;
                         }
                       next_line_num ();
                     }
-                  if (fwrite (buf, 1, buffer + BUFSIZ - buf, stdout) < buffer + BUFSIZ - buf)
+                  if (full_write (STDOUT_FILENO, buf, bytes_to_write) != bytes_to_write)
                     {
                       error (0, errno, _("write error"));
                       return false;
                     }
-                  bytes_to_write += buffer + BUFSIZ - buf;
                   seen_complete_line = false;
                   break;
                }
@@ -364,7 +362,7 @@ range_lines (const char *filename, int fd)
   int i;
   long line_from, line_to;
   off_t offset_from, offset_to;
-  
+
   ok = true;
 
   if (print_headers)
@@ -391,14 +389,17 @@ determine_seek_offsets (const char *filename, int fd)
 {
   char buffer[BUFSIZ];
   int i;
-  long line_iter, range_abs;
+  long line_iter, range_abs, range_end;
   size_t bytes_offset;
   off_t seek_offset;
 
   i = 0;
   bytes_offset = 0;
   line_iter = 0;
-  seek_offset = 0; 
+  seek_offset = 0;
+
+  range_end = ranges[range_max];
+  relative_range_to_absolute (&range_end);
 
   if ((lseek (fd, 0, SEEK_SET)) < 0)
     {
@@ -419,7 +420,7 @@ determine_seek_offsets (const char *filename, int fd)
         break;
       while (bytes_to_iter < bytes_read)
         {
-          if (buffer[bytes_to_iter] == '\n') 
+          if (buffer[bytes_to_iter] == '\n')
             {
               line_iter++;
               break;
@@ -433,6 +434,12 @@ determine_seek_offsets (const char *filename, int fd)
       if (line_iter == range_abs)
         offsets[i++] = seek_offset;
 
+      /* We don't need to determine seek offsets past the specified final
+         ending range, which is always guaranteed to be the last line.
+         This will mostly avoid a tremendous speed impact, when input is huge.  */
+      if (line_iter == range_end)
+        break;
+
       if ((seek_offset = lseek (fd, bytes_offset, SEEK_SET)) < 0)
         {
           error (0, errno, _("cannot lseek %s"), quote (filename));
@@ -519,7 +526,7 @@ validate_ranges (const char *filename, int fd)
 static void
 exit_on_zero_read (const char *filename, int fd)
 {
-  char buffer[BUFSIZ];  
+  char buffer[BUFSIZ];
   size_t bytes_read = safe_read (fd, buffer, BUFSIZ);
 
   if (!bytes_read)
@@ -553,7 +560,7 @@ buffer_stdin_to_tmpfile (const char *filename, int fd)
         }
       if (bytes_read == 0)
         break;
-      if (fwrite (buffer, 1, bytes_read, tempfile) < bytes_read)
+      if (full_write (fileno (tempfile), buffer, bytes_read) != bytes_read)
         {
           error (0, errno, _("write error"));
           return false;
@@ -659,7 +666,7 @@ parse_range_sets (char *ranges_arg)
               range_digits++;
             }
           digits_count++;
-          buf = xcalloc (digits_count, sizeof (char));
+          buf = xmalloc (digits_count * sizeof (char));
           buf_orig = buf;
           while (VALID_RANGE (*ranges_arg))
             {
@@ -677,7 +684,10 @@ parse_range_sets (char *ranges_arg)
                   ranges[range_max] = ranges[range_max - 1] + digit_increment;
                 }
               else
-                error (EXIT_FAILURE, 0, _("+ prefix can not be applied to starting lines"));
+                {
+                  free (buf); /* Would otherwise result in a leak, because a fatal error is being raised. */
+                  error (EXIT_FAILURE, 0, _("+ prefix cannot be applied to starting lines"));
+                }
             }
           else
             ranges[range_max] = atol (buf);
@@ -718,10 +728,12 @@ parse_range_sets (char *ranges_arg)
      }
   if (insert_range_at_end)
     ranges[range_max] = -1;
+  else
+    range_max--;
 }
 
 int
-main(int argc, char **argv) 
+main(int argc, char **argv)
 {
   char *ranges_arg;
   enum header_mode header_mode = multiple_files;
